////////////////////////////////////////////////////////////////////////////////
//                                                                             /
// 2012-2020 (c) Baical                                                        /
//                                                                             /
// This library is free software; you can redistribute it and/or               /
// modify it under the terms of the GNU Lesser General Public                  /
// License as published by the Free Software Foundation; either                /
// version 3.0 of the License, or (at your option) any later version.          /
//                                                                             /
// This library is distributed in the hope that it will be useful,             /
// but WITHOUT ANY WARRANTY; without even the implied warranty of              /
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           /
// Lesser General Public License for more details.                             /
//                                                                             /
// You should have received a copy of the GNU Lesser General Public            /
// License along with this library.                                            /
//                                                                             /
////////////////////////////////////////////////////////////////////////////////
#ifndef UP7_CONTEXT_H
#define UP7_CONTEXT_H

struct stuP7context                                    
{
    bool                    bInitialized;         /**< initialized flag */
    uint32_t                uSessionId;           /**< Session ID*/  
    uint8_t                 bCpuId;               /**< cpu id */
    void                   *pCtxTimer;            /**< Context for timer functions */
    fnuP7getTimerFrequency  cbTimerFrequency;     /**< callback to get system high-resolution timer frequency */
    fnuP7getTimerValue      cbTimerValue;         /**< callback to get system high-resolution timer value */
                                                  
    void                   *pCtxMem;              /**< Context for memory functions */
    fnuP7malloc             cbMalloc;             /**< callback to allocate memory */
    fnuP7free               cbFree;               /**< callback to free memory */
                                                   
    void                   *pCtxLock;             /**< Context for lock functions, set it to NULL if there is no RTOS*/
    fnuP7lock               cbLock;               /**< callback to call OS lock function, set it to NULL if there is no RTOS */
    fnuP7unlock             cbUnLock;             /**< callback to call OS unlock function, set it to NULL if there is no RTOS */
                                                  
    void                   *pCtxPacket;           /**< Context for data sending (cbSendPacket function) */
    fnuP7sendPacket         cbSendPacket;         /**< callback to send data packet to host */  

    void                   *pCtxThread;           /**< Context for cbGetCurrentThreadId function */
    fnuP7getCurrentThreadId cbGetCurrentThreadId; /**< callback get current thread ID, set it to NULL if there is no RTOS*/  

    enum euP7Level          eDefaultVerbosity;    /**< Default verbosity*/  

    struct stuTree         *pTrcModules;          /**< Trace modules map*/  
    struct stuTree         *pTrcDesc;             /**< Trace descriptors map*/  
    struct stuTree         *pTelDesc;             /**< Telemetry descriptors map*/  

    uint32_t                uTrcCounter;          /**< Trace items counter*/  

    uint64_t                qwCreationTime;       /**< Creation time*/  

    struct stP7packetChunk  stTelChunk;           /**< Cached telemetry sample chunk*/  
    stuP7telHdr             stTelSample;          /**< Cached telemetry sample value*/

    struct stuP7Module     *pModules;             /**< Trace modules descriptors, generated by uP7 preprocessor*/  
    size_t                  szModules;            /**< Trace modules descriptors count*/  

    struct stuP7Trace      *pTraces;              /**< Trace descriptors, generated by uP7 preprocessor*/  
    size_t                  szTraces;             /**< Trace descriptors count*/  

    struct stuP7telemetry  *pTelemetry;           /**< Telemetry descriptors, generated by uP7 preprocessor*/  
    size_t                  szTelemetry;          /**< Telemetry descriptors count*/  
};

#endif //UP7_CONTEXT_H
